
#include <windows.h>
#include <time.h>
#include <dshow.h>
#pragma comment(lib,"strmiids.lib")

#define MAX_FILTERS 10
#define MAX_STRING  1024

IBaseFilter   *g_source_curr = NULL;
IBaseFilter   *g_source_next = NULL;
IGraphBuilder *g_graph       = NULL;
IMediaControl *g_contr       = NULL;
IMediaSeeking *g_seekg       = NULL;
IMediaEvent   *g_event       = NULL;
HANDLE         g_event_play  = 0;
HANDLE         g_event_stop  = 0;
HANDLE         g_event_quit  = 0;
HANDLE         g_thread      = 0;
BOOL           g_init        = FALSE;
WCHAR          g_track[MAX_STRING];

//------------------------------------------------------------------------------
// This routine is used to change the source file in the current graph.
// First the graph is stopped, then the current source filter is removed.
// The new source filter is added, the output pin on this filter is
// rendered, and playback is restarted.
//------------------------------------------------------------------------------
void swap_source_filter()
{
	IPin         *pin;
	IEnumFilters *filter_enum;
	IBaseFilter  *filters[MAX_FILTERS];
	IBaseFilter **ptr;
	int           num  = 0;
	int           pos  = 0;
	LONGLONG      pos0 = 0;

	// Add the new source filter to the graph.
	if(FAILED(g_graph->AddSourceFilter(g_track, g_track, &g_source_next)))
		return;

	// Get the first output pin of the new source filter.
	if(FAILED(g_source_next->FindPin(L"Output", &pin)))
		return;

	g_contr->Stop();

	// Break all connections on the filters. You can do this
	// by adding and removing each filter in the graph.
	if(FAILED(g_graph->EnumFilters(&filter_enum)))
		return;

	// Check how many filters exists.
	while(filter_enum->Skip(1) == S_OK)
		num++;

	// Allocate space.
	if(num <= MAX_FILTERS)
		ptr = filters;
	else if(!(ptr = new IBaseFilter* [num]))
		return;

	filter_enum->Reset();
	while(filter_enum->Next(1, &(ptr[pos++]), NULL) == S_OK){}
	filter_enum->Release();

	for(pos = 0; pos < num; pos++)
	{
		g_graph->RemoveFilter(ptr[pos]);

		// Put the filter back, unless it is the old source.
		if(ptr[pos] != g_source_curr)
			g_graph->AddFilter(ptr[pos], NULL);

		ptr[pos]->Release();
	}
	if(num > MAX_FILTERS)
		delete [] ptr;

	// We have the new ouput pin. Render it.
	g_graph->Render(pin);
	g_source_curr = g_source_next;
	g_source_next = NULL;
	pin->Release();

	// Re-seek the graph to the beginning
	g_seekg->SetPositions(&pos0, AM_SEEKING_AbsolutePositioning,
						  &pos0, AM_SEEKING_NoPositioning);

	// Start the graph
	g_contr->Run();

	// Release the old source filter.
	g_source_curr->Release();
}

//------------------------------------------------------------------------------
// The secondary thread's function.
//------------------------------------------------------------------------------
void WINAPI thread_func(void *nothingrallyimportant)
{
	HANDLE   events[4];
	DWORD    wait_result;
	LONG     code, param1, param2;
	LONGLONG pos0 = 0;

	g_event->GetEventHandle((OAEVENT*)&events[0]);
	events[1] = g_event_play;
	events[2] = g_event_stop;
	events[3] = g_event_quit;

	while(1)
	{
		wait_result = WaitForMultipleObjects(4, events, FALSE, INFINITE);

		// Message from DirectShow. We are interested only in a stop message.
		// When it occurs, function resets music track to the beginning.
		if(wait_result == WAIT_OBJECT_0)
		{
			if(SUCCEEDED(g_event->GetEvent(&code, &param1, &param2, 0)))
			{
				g_event->FreeEventParams(code, param1, param2);
				if(code == EC_COMPLETE)
					g_seekg->SetPositions(&pos0, AM_SEEKING_AbsolutePositioning,
										  &pos0, AM_SEEKING_NoPositioning);
			}
		}

		// Message from user, telling that we must play other track.
		else if(wait_result == WAIT_OBJECT_0+1)
		{
			ResetEvent(g_event_play);
			swap_source_filter();
		}

		// Message from user, telling that we must stop playing.
		else if(wait_result == WAIT_OBJECT_0+2)
		{
			ResetEvent(g_event_stop);
			g_contr->Stop();
		}

		// Message from user, telling that we must quit.
		else if(wait_result == WAIT_OBJECT_0+3)
			ExitThread(0);
	}
}

//------------------------------------------------------------------------------
// Library shutdown.
//------------------------------------------------------------------------------
void shutdown()
{
	if(g_thread)
	{
		SetEvent(g_event_quit);
		WaitForSingleObject(g_thread, INFINITE);
		CloseHandle(g_thread);
	}
	if(g_event_quit) CloseHandle(g_event_quit);
	if(g_event_stop) CloseHandle(g_event_stop);
	if(g_event_play) CloseHandle(g_event_play);
	if(g_contr)      g_contr->Stop();
	if(g_event)      g_event->Release();
	if(g_seekg)      g_seekg->Release();
	if(g_contr)      g_contr->Release();
	if(g_graph)      g_graph->Release();
	CoUninitialize();
	g_init = FALSE;
}

//------------------------------------------------------------------------------
// Library initiation.
//------------------------------------------------------------------------------
BOOL initiation()
{
	if(FAILED(CoInitialize(NULL)))
	{
		shutdown();
		return FALSE;
	}
	if(FAILED(CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
		IID_IGraphBuilder, (void **)&g_graph)))
	{
		shutdown();
		return FALSE;
	}
	if(FAILED(g_graph->QueryInterface(IID_IMediaControl, (void **)&g_contr)))
	{
		shutdown();
		return FALSE;
	}
	if(FAILED(g_graph->QueryInterface(IID_IMediaSeeking, (void **)&g_seekg)))
	{
		shutdown();
		return FALSE;
	}
	if(FAILED(g_graph->QueryInterface(IID_IMediaEvent, (void **)&g_event)))
	{
		shutdown();
		return FALSE;
	}
	if(!(g_event_play = CreateEvent(NULL, TRUE, FALSE, NULL)))
	{
		shutdown();
		return FALSE;
	}
	if(!(g_event_stop = CreateEvent(NULL, TRUE, FALSE, NULL)))
	{
		shutdown();
		return FALSE;
	}
	if(!(g_event_quit = CreateEvent(NULL, TRUE, FALSE, NULL)))
	{
		shutdown();
		return FALSE;
	}
	if(!(g_thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)thread_func,
		NULL, 0, NULL)))
	{
		shutdown();
		return FALSE;
	}
	srand(time(0));
	g_init = TRUE;
	return TRUE;
}

//------------------------------------------------------------------------------
// Exported function.
//------------------------------------------------------------------------------
int __stdcall mp3play(const char *track)
{
	// Check if library is initiated.
	if(!g_init && !initiation())
		return(0);

	// Send a message to play track.
	else if(track)
	{
		char data[MAX_STRING];
		char *ptr = data;

		while(*ptr++ = *track++){}
		long length = ptr - data - 1;

//		data[length-4] = '.';
//		data[length-3] = 'm';
		data[length-2] = 'p';
		data[length-1] = '3';

		// HACK: If file doesn't exist, check if there are files with
		// the same name, but terminated with "_1", "_2", and so on.
		if(GetFileAttributes(data) == (DWORD)-1)
		{
			data[length-4] = '_';
//			data[length-3] = '1';
			data[length-2] = '.';
			data[length-1] = 'm';
			data[length]   = 'p';
			data[length+1] = '3';
			data[length+2] = '\0';

			for(int index = 0; index < 8; index++)
			{
				data[length-3] = '1' + (char)index;
				if(GetFileAttributes(data) == (DWORD)-1)
					break;
			}
			if(!index)
				return(0);

			// If there are more than one file, we select random file.
			index = rand() % index;
			data[length-3] = '1' + (char)index;
		}

		MultiByteToWideChar(CP_ACP, 0, data, -1, g_track, MAX_STRING);
		SetEvent(g_event_play);

		// Maybe we should wait until threadfunc return some value, but we
		// want to give back control to the user as fast as it is possible.
		return(1);
	}

	// Send a message to stop playing.
	else
	{
		SetEvent(g_event_stop);
        return(1);
	}
}

//------------------------------------------------------------------------------
// Main library function.
//------------------------------------------------------------------------------
BOOL APIENTRY DllMain(HANDLE module, DWORD reason, LPVOID reserved)
{
	if(reason == DLL_PROCESS_DETACH)
		shutdown();

    return TRUE;
}
